"""Library implementing the "Reader" functor."""

from __future__ import annotations

import builtins
from collections.abc import Iterable, Sequence
from dataclasses import is_dataclass
import functools
import operator as ops
from typing import Any, Callable, Generic, Optional, TypeVar


__version__ = '0.1.1'


S = TypeVar('S')
A = TypeVar('A')
B = TypeVar('B')


####################
# HELPER FUNCTIONS #
####################


def _get_attr_type(cls: type, attr: str) -> Optional[type]:
    """Given a type, if it is a dataclass or namedtuple type, checks whether it possesses the given attribute.
    If not, raises a TypeError. Otherwise, returns the type of the attribute.
    If the input type is not a dataclass or nameduple type, returns None."""
    if isinstance(cls, type):
        if is_dataclass(cls):
            if (field := cls.__dataclass_fields__.get(attr)) is not None:
                return field.type  # type: ignore[return-value]
            raise TypeError(f'invalid field {attr!r} for type {cls.__name__!r}')
        if hasattr(cls, '_fields'):  # assume it's a namedtuple
            fields = set(cls._fields)
            if attr in fields:
                # may or may not have an annotation
                return cls.__annotations__.get(attr)
            raise TypeError(f'invalid field {attr!r} for type {cls.__name__!r}')
    return None


def _getattr_simple(attr: str) -> Callable[[Any], Any]:
    assert isinstance(attr, str)
    if (not attr.isalnum()) or attr[0].isdigit():
        raise ValueError(f'invalid attribute name {attr!r}')
    return ops.attrgetter(attr)


def _getattr_nested(attrs: Sequence[str], type: Optional[type] = None) -> Callable[[Any], Any]:
    assert len(attrs) > 0
    attr = attrs[0]
    outer = _getattr_simple(attr)
    # validate attribute statically, if applicable
    inner_type = _get_attr_type(type, attr)  # type: ignore[arg-type]
    if len(attrs) == 1:
        return outer
    # recursively call this function to get a function for the nested accesses
    inner = _getattr_nested(attrs[1:], type=inner_type)
    # left-compose inner with outer
    return lambda val: inner(outer(val))


def _getattr(attr: str, *args: Any, type: Optional[type] = None) -> Callable[[Any], Any]:
    if not isinstance(attr, str):
        raise TypeError('attr must be a string')
    attrs = attr.split('.')
    func = _getattr_nested(attrs, type=type)
    if args:
        if len(args) > 1:
            num_args = 1 + len(args)
            raise TypeError(f'getattr expected at most 2 arguments, got {num_args}')
        def _getattr(val: Any) -> Any:
            try:
                return func(val)
            except AttributeError:  # return the default
                return args[0]
        return _getattr
    else:  # no default
        return func


##########
# READER #
##########

class Reader(Generic[S, A]):
    """Class that wraps a function func : S -> A."""

    def __init__(self, func: Callable[[S], A]) -> None:
        self.func = func

    @classmethod
    def const(cls, val: A) -> Reader[S, A]:
        """Given a value, returns a Reader that is a constant function returning that value."""
        return Reader(lambda _: val)

    @classmethod
    def make_tuple(cls, *readers: Reader[S, A]) -> Reader[S, tuple[A, ...]]:
        """Converts multiple Readers into a single Reader that produces a tuple of values, one for each of the input Readers."""
        return Reader(lambda val: tuple(reader(val) for reader in readers))

    def __call__(self, val: S) -> A:
        """Calls the wrapped function."""
        return self.func(val)

    def map(self, func: Callable[[A], B]) -> Reader[S, B]:
        """Left-composes a function onto the wrapped function, returning a new Reader."""
        return Reader(lambda val: func(self(val)))

    def map_binary(self, operator: Callable[[A, A], B], other: Reader[S, A]) -> Reader[S, B]:
        """Given a binary operator and another Reader, returns a new Reader that applies the operator to the output of this Reader and the other Reader."""
        return Reader(lambda val: operator(self(val), other(val)))

    # ARITHMETIC OPERATORS

    def __add__(self, other: Reader[S, A]) -> Reader[S, A]:
        return self.map_binary(ops.add, other)

    def __sub__(self, other: Reader[S, A]) -> Reader[S, A]:
        return self.map_binary(ops.sub, other)

    def __mul__(self, other: Reader[S, A]) -> Reader[S, A]:
        return self.map_binary(ops.mul, other)

    def __truediv__(self, other: Reader[S, A]) -> Reader[S, A]:
        return self.map_binary(ops.truediv, other)

    def __mod__(self, other: Reader[S, A]) -> Reader[S, A]:
        return self.map_binary(ops.mod, other)

    def __floordiv__(self, other: Reader[S, A]) -> Reader[S, A]:
        return self.map_binary(ops.floordiv, other)

    def __pow__(self, other: Reader[S, A]) -> Reader[S, A]:
        return self.map_binary(ops.pow, other)

    def __matmul__(self, other: Reader[S, A]) -> Reader[S, A]:
        return self.map_binary(ops.matmul, other)

    def __neg__(self) -> Reader[S, A]:
        return self.map(ops.neg)  # type: ignore[arg-type]

    def __pos__(self) -> Reader[S, A]:
        return self.map(ops.pos)  # type: ignore[arg-type]

    def __invert__(self) -> Reader[S, A]:
        return self.map(ops.inv)  # type: ignore[arg-type]

    # LOGICAL OPERATORS

    def __and__(self, other: Reader[S, A]) -> Reader[S, A]:
        return self.map_binary(ops.and_, other)

    def __or__(self, other: Reader[S, A]) -> Reader[S, A]:
        return self.map_binary(ops.or_, other)

    def __xor__(self, other: Reader[S, A]) -> Reader[S, A]:
        return self.map_binary(ops.xor, other)

    def truthy(self) -> Reader[S, bool]:
        """Returns a Reader that evaluates the `bool` function on this Reader's output."""
        return self.map(bool)

    def falsy(self) -> Reader[S, bool]:
        """Returns a Reader that evaluates the logical negation (`not` operator) on this Reader's output."""
        return self.map(ops.not_)

    # COMPARISON OPERATORS

    def __lt__(self, other: Reader[S, A]) -> Reader[S, bool]:
        return self.map_binary(ops.lt, other)  # type: ignore[arg-type]

    def __le__(self, other: Reader[S, A]) -> Reader[S, bool]:
        return self.map_binary(ops.le, other)  # type: ignore[arg-type]

    def __ge__(self, other: Reader[S, A]) -> Reader[S, bool]:
        return self.map_binary(ops.ge, other)  # type: ignore[arg-type]

    def __gt__(self, other: Reader[S, A]) -> Reader[S, bool]:
        return self.map_binary(ops.gt, other)  # type: ignore[arg-type]

    def equals(self, other: Reader[S, A]) -> Reader[S, bool]:
        """Returns a Reader that evaluates whether the output of this Reader equals the output of the other."""
        return self.map_binary(ops.eq, other)

    def not_equals(self, other: Reader[S, A]) -> Reader[S, bool]:
        """Returns a Reader that evaluates whether the output of this Reader does not equal the output of the other."""
        return self.equals(other).falsy()

    # OTHER OPERATORS

    def contains(self, element: Any) -> Reader[S, bool]:
        """Returns a Reader returning True if the given element is in the value returned by this Reader."""
        return self.map(lambda val: element in val)  # type: ignore[operator]

    def getitem(self, index: Any) -> Reader[S, Any]:
        """Returns a Reader that returns value[index], where value is the value returned by this Reader."""
        return self.map(ops.itemgetter(index))  # type: ignore[arg-type]

    def getattr(self, attr: str, *args: Any, type: Optional[type] = None) -> Reader[S, Any]:
        """Returns a Reader that returns getattr(value, attr, [default]), where value is the value returned by this Reader.
        attr may contain multiple fields separated by '.' (example x.y.z), which will perform nested attribute retrieval.
        The second argument to this method, if present, is the default value to use if an attribute does not exist."""
        return self.map(_getattr(attr, *args, type=type))


#######################
# READER CONSTRUCTORS #
#######################


def const(val: A) -> Reader[S, A]:
    """Given a value, returns a Reader that is a constant function returning that value."""
    return Reader.const(val)


def make_tuple(*readers: Reader[S, A]) -> Reader[S, tuple[A, ...]]:
    """Converts multiple Readers into a single Reader that produces a tuple of values, one for each of the input Readers."""
    return Reader.make_tuple(*readers)


def itemgetter(index: Any) -> Reader[S, Any]:
    """Given an index object, returns a Reader that takes a value and returns returns value[index]."""
    return Reader(ops.itemgetter(index))  # type: ignore[arg-type]


def attrgetter(attr: str, *args: Any, type: Optional[type] = None) -> Reader[S, Any]:
    """Given an attribute string and optional default, returns a Reader that takes a value and returns returns getattr(value, attr, [default]).
    attr may contain multiple fields separated by '.' (example x.y.z), which will perform nested attribute retrieval."""
    return Reader(_getattr(attr, *args, type=type))


def map(func: Callable[[A], B], reader: Reader[S, A]) -> Reader[S, B]:  # noqa: A001
    """Left composes a function onto a Reader, returning a new Reader."""
    return reader.map(func)


##############
# REDUCTIONS #
##############


def reduce(readers: Iterable[Reader[S, A]], operator: Callable[[A, A], A], initial: Optional[A] = None) -> Reader[S, A]:
    """Given a sequence of Readers and a binary operator, produces a new Reader that reduces the operator over the values produced by the input Readers.
    An initial value can optionally be provided to handle the case where an empty sequence is acted on."""
    if initial is None:
        func = functools.partial(functools.reduce, operator)
    else:
        func = lambda iterable: functools.reduce(operator, iterable, initial)  # type: ignore[assignment]
    return make_tuple(*readers).map(func)


def all(readers: Iterable[Reader[S, A]]) -> Reader[S, bool]:  # noqa: A001
    """Given a sequence of Readers, produces a new Reader that evaluates the `all` function over the values output by the Readers."""
    return reduce(readers, ops.and_, initial=True)  # type: ignore[arg-type]


def any(readers: Iterable[Reader[S, A]]) -> Reader[S, bool]:  # noqa: A001
    """Given a sequence of Readers, produces a new Reader that evaluates the `any` function over the values output by the Readers."""
    return reduce(readers, ops.or_, initial=False)  # type: ignore[arg-type]


def sum(readers: Iterable[Reader[S, A]], start: Optional[A] = None) -> Reader[S, A]:  # noqa: A001
    """Given a sequence of Readers, produces a new Reader that evaluates the sum of the values output by the Readers."""
    return reduce(readers, ops.add, initial=start)


def prod(readers: Iterable[Reader[S, A]], start: Optional[A] = None) -> Reader[S, A]:
    """Given a sequence of Readers, produces a new Reader that evaluates the product of the values output by the Readers."""
    return reduce(readers, ops.mul, initial=start)


def min(readers: Iterable[Reader[S, A]], default: Optional[A] = None) -> Reader[S, A]:  # noqa: A001
    """Given a sequence of Readers, produces a new Reader that evaluates the minimum of the values output by the Readers."""
    return reduce(readers, builtins.min, initial=default)  # type: ignore[arg-type]


def max(readers: Iterable[Reader[S, A]], default: Optional[A] = None) -> Reader[S, A]:  # noqa: A001
    """Given a sequence of Readers, produces a new Reader that evaluates the maximum of the values output by the Readers."""
    return reduce(readers, builtins.max, initial=default)  # type: ignore[arg-type]



from collections import namedtuple
from dataclasses import dataclass
import operator as ops
import re
from typing import Any, NamedTuple

import pytest

import undercat as uc
from undercat import Reader


def square(x):
    return x ** 2

r_square = Reader(square)

def add_one(x):
    return x + 1

r_add_one = Reader(add_one)

r_id = Reader(lambda val: val)
r_not = Reader(ops.not_)
r_item1 = uc.itemgetter(1)
r_attr = uc.attrgetter('attr')


class Vec(tuple[float]):
    """Class for a vector of floats that implements uses the dot product as the @ operator."""

    def __new__(cls, *args):
        return super().__new__(cls, args)

    def __matmul__(self, other):
        return sum(xi * yi for (xi, yi) in zip(self, other))


@dataclass
class Obj:
    attr: Any

r_attr_typed = uc.attrgetter('attr', type=Obj)

class Outer:
    def __init__(self, val1):
        self.val1 = val1

@dataclass
class OuterDC:
    val1: Obj
    val2: int

class OuterTypedNT(NamedTuple):
    val1: Obj
    val2: int

OuterUntypedNT = namedtuple('OuterUntypedNT', ['val1', 'val2'])


@pytest.mark.parametrize(['reader', 'input_val', 'output_val'], [
    # Reader
    (r_square, 3, 9),
    (r_add_one, 0, 1),
    # const
    (uc.const(5), 0, 5),
    (uc.const(5), 1, 5),
    # itemgetter
    (r_item1, 3, TypeError('not subscriptable')),
    (r_item1, [], IndexError('out of range')),
    (r_item1, [1, 2, 3], 2),
    (uc.itemgetter('key'), {}, KeyError('key')),
    (uc.itemgetter('key'), {'key': 3}, 3),
    # attrgetter
    (r_attr, 3, AttributeError("'int' object has no attribute 'attr'")),
    (r_attr_typed, 3, AttributeError("'int' object has no attribute 'attr'")),
    (r_attr, Obj(3), 3),
    (r_attr_typed, Obj(3), 3),
    (uc.attrgetter('fake'), Obj(3), AttributeError("'Obj' object has no attribute 'fake'")),
    (uc.attrgetter('attr.attr'), Obj(Obj(3)), 3),
    (uc.attrgetter('fake', type=Outer), Outer(Obj(3)), AttributeError("'Outer' object has no attribute 'fake'")),
    (uc.attrgetter('val1.attr', type=Outer), Outer(Obj(3)), 3),
    (uc.attrgetter('val1.fake', type=Outer), Outer(Obj(3)), AttributeError("'Obj' object has no attribute 'fake'")),
    (uc.attrgetter('val1.attr', type=OuterDC), OuterDC(Obj(3), 1), 3),
    (uc.attrgetter('val2.attr', type=OuterDC), OuterDC(Obj(3), 1), AttributeError("'int' object has no attribute 'attr'")),
    (uc.attrgetter('val1.attr', type=OuterTypedNT), OuterDC(Obj(3), 1), 3),
    (uc.attrgetter('val1.attr', type=OuterUntypedNT), OuterDC(Obj(3), 1), 3),
    (uc.attrgetter('val1.fake', type=OuterUntypedNT), OuterDC(Obj(3), 1), AttributeError("'Obj' object has no attribute 'fake'")),
    # make_tuple
    (uc.make_tuple(uc.const(1), uc.const(2)), 0, (1, 2)),
    # map
    (r_square.map(add_one), 3, 10),
    (uc.map(add_one, r_square), 3, 10),
    (r_add_one.map(square), 3, 16),
    (uc.map(square, r_add_one), 3, 16),
    (uc.const(5).map(square), 0, 25),
    # map_binary
    (r_add_one.map_binary(ops.add, r_square), 3, 13),
    # arithmetic operators
    (-r_square, 3, -9),
    (+r_square, 3, 9),
    (~r_square, 3, -10),
    (r_square + r_add_one, 3, 13),
    (r_square - r_add_one, 3, 5),
    (r_square * r_add_one, 3, 36),
    (r_square / r_add_one, 3, 2.25),
    (r_square % r_add_one, 3, 1),
    (r_square // r_add_one, 3, 2),
    (r_square ** r_add_one, 3, 9 ** 4),
    (uc.const(Vec(1, 2)) @ uc.const(Vec(3, 4)), None, 11),
    # logical operators
    (uc.const(True).truthy(), 0, True),
    (uc.const(False).truthy(), 1, False),
    (uc.const(True).falsy(), 0, False),
    (uc.const(False).falsy(), 0, True),
    (r_id.falsy(), True, False),
    (r_id.falsy(), False, True),
    (r_not.falsy(), True, True),
    (r_not.falsy(), False, False),
    (uc.const(3) & uc.const(2), 0, 2),
    (uc.const(3) and uc.const(2), 0, 2),
    (uc.const(True) & uc.const(False), 0, False),
    (uc.const(True) & uc.const(True), 0, True),
    (uc.const(True) and uc.const(False), 0, False),
    (uc.const(True) and uc.const(True), 0, True),
    (r_not and uc.const(True), False, True),
    (r_not and r_not, False, True),
    (uc.const(3) | uc.const(2), 0, 3),
    (uc.const(3) or uc.const(2), 0, 3),
    (uc.const(True) | uc.const(False), 0, True),
    (uc.const(False) | uc.const(False), 0, False),
    (uc.const(True) or uc.const(False), 0, True),
    (uc.const(False) or uc.const(False), 0, False),
    (uc.const(3) ^ uc.const(2), 0, 1),
    (uc.const(True) ^ uc.const(False), 0, True),
    (uc.const(False) ^ uc.const(False), 0, False),
    (uc.const(True) ^ uc.const(True), 0, False),
    # comparison operators
    (r_square < r_add_one, 3, False),
    (r_square < r_square, 3, False),
    (r_square <= r_add_one, 3, False),
    (r_square <= r_square, 3, True),
    (r_square >= r_add_one, 3, True),
    (r_square >= r_square, 3, True),
    (r_square > r_add_one, 3, True),
    (r_square > r_square, 3, False),
    (r_square.equals(r_square), 3, True),
    (r_square.equals(r_id), 3, False),
    (r_square.not_equals(r_square), 3, False),
    (r_square.not_equals(r_id), 3, True),
    # other operators
    (r_id.contains('a'), 3, TypeError('not iterable')),
    (r_id.contains('a'), 'abc', True),
    (r_id.contains('a'), 'bc', False),
    (r_id.contains('a'), ['a'], True),
    (r_id.contains('a'), ['ab'], False),
    (r_id.getitem(0), 3, TypeError('not subscriptable')),
    (r_id.getitem(0), [], IndexError('out of range')),
    (r_id.getitem(0), [1, 2, 3], 1),
    (r_id.getitem(slice(None, 2)), 3, TypeError('not subscriptable')),
    (r_id.getitem(slice(None, 2)), [], []),
    (r_id.getitem(slice(None, 2)), [1, 2, 3], [1, 2]),
    (r_id.getitem('key'), 3, TypeError('not subscriptable')),
    (r_id.getitem('key'), [], TypeError('list indices must be integers or slices')),
    (r_id.getitem('key'), {}, KeyError('key')),
    (r_id.getitem('key'), {'key': 'val'}, 'val'),
    (r_id.getattr('attr'), 3, AttributeError("'int' object has no attribute 'attr'")),
    (r_id.getattr('attr'), Obj(3), 3),
    (r_id.getattr('attr', None), Obj(3), 3),
    (r_id.getattr('other', None), Obj(3), None),
    (r_id.getattr('attr'), Obj(Obj(3)), Obj(3)),
    (r_id.getattr('attr').getattr('attr'), Obj(3), AttributeError("'int' object has no attribute 'attr'")),
    (r_id.getattr('attr').getattr('attr'), Obj(Obj(3)), 3),
    (r_id.getattr('attr.attr'), 3, AttributeError("'int' object has no attribute 'attr'")),
    (r_id.getattr('attr.attr'), Obj(3), AttributeError("'int' object has no attribute 'attr'")),
    (r_id.getattr('attr.attr'), Obj(Obj(3)), 3),
    (r_id.getattr('attr.attr', None), 3, None),
    (r_id.getattr('attr.attr', None), Obj(3), None),
    (r_id.getattr('attr.attr', None), Obj(Obj(3)), 3),
    # reductions
    (uc.all([]), False, True),
    (uc.all([]), True, True),
    (uc.all([r_id, r_id, r_id]), False, False),
    (uc.all([r_id, r_id, r_id]), True, True),
    (uc.all([r_id, r_not, r_id]), False, False),
    (uc.all([r_id, r_not, r_id]), True, False),
    (uc.any([]), False, False),
    (uc.any([]), True, False),
    (uc.any([r_id, r_id, r_id]), False, False),
    (uc.any([r_id, r_id, r_id]), True, True),
    (uc.any([r_id, r_not, r_id]), False, True),
    (uc.any([r_id, r_not, r_id]), True, True),
    (uc.sum([]), None, TypeError('no initial value')),
    (uc.sum([r_id, r_square, r_add_one]), 3, 16),
    (uc.sum([uc.const(1), uc.const('2')]), None, TypeError('unsupported operand type')),
    (uc.sum([uc.const('1'), uc.const('2')]), None, '12'),
    (uc.sum([uc.const([1]), uc.const([]), uc.const([2])]), None, [1, 2]),
    (uc.prod([]), None, TypeError('no initial value')),
    (uc.prod([r_id, r_square, r_add_one]), 3, 108),
    (uc.min([]), None, TypeError('no initial value')),
    (uc.min([r_id, r_square, r_add_one]), 3, 3),
    (uc.max([]), None, TypeError('no initial value')),
    (uc.max([r_id, r_square, r_add_one]), 3, 9),
])
def test_reader(reader, input_val, output_val):
    """Tests that a (reader, input) pair produces what we expect."""
    if isinstance(output_val, Exception):  # expect an error
        with pytest.raises(type(output_val), match=str(output_val)):
            _ = reader(input_val)
    else:
        assert reader(input_val) == output_val

def test_bool_operators():
    """Tests that the `bool` and `not` operators return a bool when evaluated on a Reader.
    (This may be unexpected, as one might think they return a Reader.)"""
    for reader in [uc.const(False), uc.const(True), r_id]:
        assert bool(reader) is True
        assert (not reader) is False

def test_reader_equality():
    """Tests that the `__eq__` and `__ne__` operators return a bool when evaluated on a pair of Readers.
    (This may be unexpected, as one might think they return a Reader.)
    Equality is just identity of objects."""
    assert r_square is r_square
    assert (r_square == r_square) is True
    assert (r_square != r_square) is False
    assert uc.const(1) is not uc.const(2)
    assert (uc.const(1) != uc.const(2)) is True
    assert (uc.const(1) == uc.const(2)) is False
    assert uc.const(1) is not uc.const(1)
    assert (uc.const(1) != uc.const(1)) is True
    assert (uc.const(1) == uc.const(1)) is False

def test_invalid_operators():
    """Tests that certain operators are invalid when called on a Reader."""
    type_err = lambda match: pytest.raises(TypeError, match=match)
    # __iter__
    with type_err('not iterable'):
        _ = list(r_square)  # type: ignore[call-overload]
    # __contains__
    with type_err('not iterable'):
        _ = 3 in r_square
    with type_err('not iterable'):
        _ = 3 not in r_square
    with type_err('not iterable'):
        _ = 3 not in uc.const([1, 2, 3])
    # __getitem__
    with type_err('not subscriptable'):
        _ = r_square[1]
    with type_err('not subscriptable'):
        _ = r_square['key']
    with type_err('not subscriptable'):
        _ = r_square[:2]
    # getattr with too many args
    with type_err('getattr expected at most 2 arguments, got 3'):
        _ = r_id.getattr('attr', 1, None)
    # getattr with invalid attr string
    with type_err('attr must be a string'):
        _ = r_id.getattr(3)

@pytest.mark.parametrize(['attr', 'seg'], [
    ('', ''),
    (' ', ' '),
    ('+', '+'),
    ('a+', 'a+'),
    ('a ', 'a '),
    (' a', ' a'),
    ('1', '1'),
    ('1a', '1a'),
    ('a.b+', 'b+'),
    ('a.1b', '1b'),
    ('a.1b.c', '1b'),
])
def test_invalid_attr_name(attr, seg):
    """Tests attrgetter when the attribute name is invalid."""
    with pytest.raises(ValueError, match=re.escape(f'invalid attribute name {seg!r}')):
        _ = uc.attrgetter(attr)

def test_invalid_getattr_typed():
    """Tests attrgetter when the provided type is invalid."""
    err = "invalid field 'fake' for type 'Obj'"
    for (attr, tp) in [('fake', Obj), ('val1.fake', OuterDC), ('val1.fake', OuterTypedNT)]:
        with pytest.raises(TypeError, match=err):
            _ = uc.attrgetter(attr, type=tp)
    for tp in [OuterDC, OuterTypedNT, OuterUntypedNT]:
        with pytest.raises(TypeError, match=f"invalid field 'fake' for type '{tp.__name__}'"):
            _ = uc.attrgetter('fake', type=tp)
